#include "Bkgreg.h"
#include <errstr.h>
#include <cmath>
#include <algorithm>
#include <limits>

Bkgreg::Bkgreg(double x, double y, int ccdnr, double r, std::string instr,std::string mode)
: _rawx(x), _rawy(y), _ccdnr(ccdnr), _rad(r), _instr(instr), _mode(mode)
{
	calReg();
}

void Bkgreg::calReg()
{
	/// MOS
	if (_instr == "EMOS1" || _instr == "EMOS2")
		calRegMOS();
	// PN
	else if (_instr == "EPN")
		calRegPN();
}

void Bkgreg::calRegMOS()
{
	std::cout << std::endl;
	errstr << "Input image instrument: "<< _instr << ". This task is only valid for Epic-PN observations. Exiting."<< Sas::error("invalidModeType");
}

void Bkgreg::calRegPN()
{
	// Source position and src extraction radius in RAWX,RAWY coordinates
	//int _rawx; int _rawy; double _rad;
	double xs = _rawx;
	double ys = _rawy;

	// Find CCD limits: xmin, xmax, ymin, ymax
	std::vector<int> ccd_limits = calCCDlimits();
	double xmin = ccd_limits[0];
	double xmax = ccd_limits[1];
	double ymin = ccd_limits[2];
	double ymax = ccd_limits[3];

	// Constants
	double pixsizePN = 4.18;
	double rb0 = 42. * 80 * 0.05;		// Initial bkg radius = 42 pix ( = pipeline)
	//double dr0 = 12.;					// Initial distance between edges of (src region - bkg region) = 12 arcsec
	double dr0 = _rad * 0.05;			// Initial distance between edges of (src region - bkg region) = Src extraction radius
		if (dr0 < 12.)					//      Clamp distance to ( 12 - 40 arcsec ) range
			dr0 = 12.;
		if (dr0 > 40.)
			dr0 = 40.;
		//--- DEBUG ---
		//std::cout << "src rad = " << _rad << ", d = " << dr0 << std::endl;
		//--- DEBUG ---

	double rs  = _rad*0.05/pixsizePN;	// Convert Sky to RAWX,RAWY
	double rb  = rb0 / pixsizePN;
	double dr  = dr0 / pixsizePN;

	// Begins bkg selection calculations
	double xsize = xmax - xmin;
	double ysize = ymax - ymin;
	double xxs = xs - xmin;
	double yys = ys - ymin;
	double rrb;
	double xswap;
	double yswap;
	double xxb;
	double yyb;
	double xb;
	double yb;

	if ( xxs < xsize/2 ){ xswap = 1; } else { xswap = 0; }
	if ( yys < ysize/2 ){ yswap = 1; } else { yswap = 0; }

	if ( xswap == 1 ){ xxs = xsize - xxs; }
	if ( yswap == 1 ){ yys = ysize - yys; }

	double rr = pow((pow(xxs,2) + pow(yys,2)), 0.5);

	//std::vector<double> distan;
	//	distan.push_back(rb);
	//	distan.push_back((xxs-dr)/2);
	//	distan.push_back((yys-rs-dr)/2);
	//	distan.push_back((rr-rs-dr)/2);
	//	std::vector<double>::const_iterator itmin =  std::min_element(distan.begin(), distan.end());
	//	rrb = *itmin;

	if (rrb < 0){ rrb = (rr-rs-dr)/(2.)/pow(2.,0.5) ; }

	std::vector<double> distan2;
		distan2.push_back(rb);
		distan2.push_back((rr-rs-dr)/(1.+pow(2.,0.5)));
		distan2.push_back((xxs-dr)/2.);
		std::vector<double>::const_iterator itmin2 =  std::min_element(distan2.begin(), distan2.end());
		rrb = *itmin2;

	if (rrb > ((xxs-rs-dr)/2.)){
		xxb = rrb;
	} else {
		xxb = ((xxs-rs-dr)/2.);
	}

	double xxs_xxb = xxs - xxb;
	if (xxs_xxb < 0){xxs_xxb = 0;}

	double sqrt_input = (1. - pow( (xxs_xxb)/(rs+dr+rrb),2. ));
	if (sqrt_input < 0){sqrt_input = 0;}

	yyb = yys - (rs+dr+rrb)*pow(sqrt_input ,0.5);

	if (yyb >= (ysize-rrb)){yyb = (ysize-rrb);}

	xxb = xxs - pow( pow((rs+dr+rrb),2.) - pow((yys-yyb),2.) , 0.5);

	if (xswap == 1){xxb = xsize-xxb;}
	if (yswap == 1){yyb = ysize-yyb;}

	xb = xxb + xmin;
	yb = yyb + ymin;

	// ==== DEBUG ====================================================================================================
	//std::cout <<"------ DEBUG ---------------------------------------"<< std::endl;
	// Input  : xs (RAWX), ys (RAWY, rs, rb, xmin, xmax, ymin, ymax, dr
	//std::cout <<"	xs, ys, rs, rb, xmin, xmax, ymin, ymax, dr = " << xs << " " << ys << " " << rs << " " << rb << " " << xmin << " " << xmax << " " << ymin << " " << ymax << " " << dr << std::endl;
	//std::cout << "" << std::endl;
	//std::cout <<"	xb, yb, rrb = " << xb << ", " << yb << ", " << rrb <<std::endl;
	//std::cout <<"------ DEBUG ---------------------------------------"<< std::endl;
	//std::cout << "" << std::endl;
	// ==== DEBUG ====================================================================================================

	if (rrb < 3){
		// If rrb < 3 pix No background extraction region is produced -> xb, yb, rrb = 0, 0, 0
			std::cout << "" << std::endl;
			errstr << "Background extraction radius too small ( < 3 pix )."<< Sas::warning("smallBkgRadius");
			std::cout << "" << std::endl;
					xb = std::numeric_limits<double>::quiet_NaN();
					yb = std::numeric_limits<double>::quiet_NaN();
					rrb = std::numeric_limits<double>::quiet_NaN();
				}

	_res.push_back(xb);
	_res.push_back(yb);
	_res.push_back(rrb);

}


std::vector<int> Bkgreg::calCCDlimits()
{
	// Min and Max values of RAWX, RAWY within the CCD:
	int xmin = 2;
	int xmax = 63;
	int ymin;
	int ymax = 199;
	if (_mode == "PrimeFullWindow" || _mode == "PrimeFullWindowExtended"){
		if (_ccdnr == 3 || _ccdnr == 6 || _ccdnr == 9 || _ccdnr == 12)
			ymin = 30;
		else
			ymin = 14;
	}
	if (_mode == "PrimeLargeWindow"){
		ymin = 103;
	}
	if (_mode == "PrimeSmallWindow"){
			ymin = 138;
	}

	ccd_limits.push_back(xmin);
	ccd_limits.push_back(xmax);
	ccd_limits.push_back(ymin);
	ccd_limits.push_back(ymax);
	return ccd_limits;
}



std::vector<double> Bkgreg::getCoord()
{
	return _res;
}
