#include "ProcSrcList.h"

#include <Tracer.h>
#include <errstr.h>
#include "param.h"
#include <cstdio>	// For tmpnam

#include "CalServer.h"
#include "Dal.h"

#include "DataFilter.h"
#include <iostream>
#include <cmath>
#include <vector>
#include <algorithm>
#include "esky2det.h"
#include "edet2sky.h"

#include "SasFileLocator.h"     //      for locating d/f color tables
#include <metatask.h>

#include "Bkgreg.h"
#include "Lookupradius.h"

using namespace std;

ProcSrcList::ProcSrcList(string src_list, string imagefile, string instr, string mode, bool withCoords, bool withSrcRad, double rad)
: _src_list(src_list), _imagefile(imagefile), _instr(instr), _mode(mode), _withCoords(withCoords), _withSrcRad(withSrcRad), _rad(rad)
{
	calProcess();
}


void ProcSrcList::calProcess()
{

	string src_table = _src_list + ":SRCLIST";
	std::cout << "Processing input Source List : " << src_table << std::endl;

	// Check previous processing by 'ebkgreg' in the input Source List
	//
	string src_table_test = _src_list + ":SRCLIST";
	DataSet *src_set_test = dataSetServer->open(src_table_test,Dal::Read);
	Table * src_tab_test = src_set_test->table("SRCLIST");
	if (src_tab_test->hasColumn("BKG_DETX") && src_tab_test->hasColumn("BKG_DETY") && src_tab_test->hasColumn("BKG_RA") &&
					src_tab_test->hasColumn("BKG_DEC") && src_tab_test->hasColumn("BKG_X") && src_tab_test->hasColumn("BKG_Y")){
			// Error message:
			errstr << "Input Source List already has background region information. Exiting."<< Sas::error("invalidSourceList");
			}
			dataSetServer->close(src_set_test);

	//
	// Create a temporal clone for working with
	//
	string temp_source_list_file = tmpnam(0);
	_res_tmpsrclist = temp_source_list_file;
	string src_table_clone = temp_source_list_file + ":SRCLIST";
	DataSet *src_set_clone = dataSetServer->clone(src_table, src_table_clone, Dal::Modify);
	dataSetServer->close(src_set_clone);
	
	// ---- DEBUG ----
	//exit(1);
	// ---- DEBUG ----
	
	//
	// Convert Sky to Detector Coordinates
	//
	// esky2det(Src RA, Dec) == Src RAWX, RAWY, CCD  --> 3 extra columns in input source list
		Info::Task* _esky2det = task("esky2det");
		ParameterInfo& esky2detpars = *_esky2det->parameters();
		esky2detpars.revertToDefault();
		esky2detpars["calinfoset"]->setValue(_imagefile);
		esky2detpars["datastyle"]->setValue("set");
		esky2detpars["intab"]->setValue(src_table_clone);

			  	try{
				   			task(esky2det);
				  		}
				  		catch (SasException::Error e) {
				  			errstr << _esky2det->name() << " terminated with the error " << e.code
				   			   << " that said: \n" << e.message
				   			   << Sas::warning("SubTaskError");
				   		}
				   		catch (SasException::Fatal e) {
				   			errstr << _esky2det->name() << " terminated with the fatal error " << e.code
				   			   << " that said: \n" << e.message
				   			   << Sas::warning("SubTaskError");
				   		}

	///////////////////////////////////////////////////////////////////////////
	//
	// Get initial extraction radius for each source from LOOKUP.FITS table

		// Read Source List 'src_set' = 'src_tab' == Input clone
		//
		DataSet *src_set = dataSetServer->open(src_table_clone,Dal::Modify);
		Table *src_tab = src_set->table("SRCLIST");
		unsigned src_N = src_tab->rows();

		// Rename RAWX, RAWY, CCDNR columns --> SRC_RAWX, SRC_RAWY, SRC_CCDNR
		//
		//  Test that the columns exist in the src_tab table before renaming
		//
		if (src_tab->hasColumn("RAWX") && src_tab->hasColumn("RAWY") && src_tab->hasColumn("CCDNR") &&
				!src_tab->hasColumn("SRC_RAWX") && !src_tab->hasColumn("SRC_RAWY") && !src_tab->hasColumn("SRC_CCDNR")){
			Column *colx = src_tab->column("RAWX");
			colx->rename("SRC_RAWX");
			Column *coly = src_tab->column("RAWY");
			coly->rename("SRC_RAWY");
			Column *colccd = src_tab->column("CCDNR");
			colccd->rename("SRC_CCDNR");

			Column *colra = src_tab->column("RA");
			colra->rename("SRC_RA");
			Column *coldec = src_tab->column("DEC");
			coldec->rename("SRC_DEC");
		}
		// If SRC_RAW columns already exist, the new columns RAWX,RAWY,CCDNR from 'esky2det' are removed
		if (src_tab->hasColumn("RAWX") && src_tab->hasColumn("RAWY") && src_tab->hasColumn("CCDNR") &&
				src_tab->hasColumn("SRC_RAWX") && src_tab->hasColumn("SRC_RAWY") && src_tab->hasColumn("SRC_CCDNR")){
			// Delete RAW columns from 'esky2det'
			src_tab->deleteColumn("RAWX");
			src_tab->deleteColumn("RAWY");
			src_tab->deleteColumn("CCDNR");
		}

		// Initial Src Extraction Radius:
		std::vector<double> init_rad;
		

		if (_withCoords){
		// If withcoords = yes ==> Initial Radius from command line option
			init_rad.push_back(_rad / 0.05);
			
		} else {
////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//
		// If withcoords = no ==> Initial Radius from (LOOKUP table) or (SRC_RAD column in the input  Src List)
		//======================================================================================================
			if (_withSrcRad){
				// Initial Radius from (SRC_RAD column in the input  Src List)
				//============================================================
				// Check input columns exist
				if (!src_tab->hasColumn("SRC_RAD")){
						// Error message:
						errstr << "Needed column not found in input Source List (SRC_RAD). Exiting."<< Sas::error("invalidSourceList");
						}
				// Ingest SRC_RAD column from input Src List to init_rad vector
				cout << "Info: Source extraction region radius from SRC_RAD" << endl;
				ColumnData *src_rad_ = src_tab->column("SRC_RAD")->data();
				real32 *src_rad = src_rad_->real32Data();
					for (unsigned long i = 0 ; i < src_N ; ++i){
						init_rad.push_back(src_rad[i] / 0.05);   // init_rad units = rad / 0.05 = arcsec * 20
					}
			}
			else {
				// Initial Radius from (LOOKUP table)
				//===================================
				// Check input columns exist
				if (!src_tab->hasColumn("PN_CTS") && !src_tab->hasColumn("PN_OFFAX")){
						// Error message:
						errstr << "Needed columns not found in input Source List (PN_CTS and PN_OFFAX). Exiting."<< Sas::error("invalidSourceList");
						}
				cout << "Info: Source extraction region radius from PN_CTS and PN_OFFAX (LOOKUP)" << endl;
				ColumnData *src_pn_cts = src_tab->column("PN_CTS")->data();
				real32 *pn_cts = src_pn_cts->real32Data();
				ColumnData *src_pn_offax = src_tab->column("PN_OFFAX")->data();
				real32 *pn_offax = src_pn_offax->real32Data();

				// Search for Radius from LOOKUP table:
				//
					for (unsigned long i = 0 ; i < src_N ; ++i){
						Lookupradius radius(pn_cts[i], pn_offax[i]);
						double rad = radius.getRadius();  // Radius from LOOKUP file (arcsec)
						init_rad.push_back(rad / 0.05);   // init_rad units = rad / 0.05 = arcsec * 20
						//cout << "--- DEBUG ---" << endl;
						//cout << "i = " << i << ", " << pn_offax[i] <<", "<< pn_cts[i]<<". Radius = "<< init_rad[i] << endl;
						//cout << "--- DEBUG ---" << endl;
					}
			}

		}
////////////////////////////////////////////////////////////////////////////////////////////////////////////


		// Get Bkg position and Bkg radius for each source
		// Bkgreg bkg(xsrc_rawx, ysrc_rawy, ccd_cent, init_rad, instr, submode);
		//
		// Src RAWX, RAWY and Src CCD from 3 extra columns of source list
		ColumnData *tab_src_rawx = src_tab->column("SRC_RAWX")->data();
		real32 *src_rawx = tab_src_rawx->real32Data();
		ColumnData *tab_src_rawy = src_tab->column("SRC_RAWY")->data();
		real32 *src_rawy = tab_src_rawy->real32Data();
		ColumnData *tab_src_ccdnr = src_tab->column("SRC_CCDNR")->data();
		int8 *src_ccdnr = tab_src_ccdnr->int8Data();

		// Call Bkg:
		std::vector<double> bkg_rawx;
		std::vector<double> bkg_rawy;
		std::vector<int8> bkg_ccdnr;  // Bkg region in the same CCD as the source
		std::vector<double> bkg_rad;
		for (unsigned long i = 0 ; i < src_N ; ++i){
			// ---- DEBUG ----
			//cout << "SRC_NUM = " << src_num[i] << ", RAW:" << src_rawx[i] <<", "<< src_rawy[i]<< ", "
			//		<< int(src_ccdnr[i]) << " . Radius = "<< init_rad[i] << endl;
			// ---- DEBUG ----

			Bkgreg bkg(src_rawx[i], src_rawy[i], src_ccdnr[i], init_rad[i], _instr, _mode);

			bkg_rawx.push_back(bkg.getCoord()[0]);
			bkg_rawy.push_back(bkg.getCoord()[1]);
			bkg_rad.push_back(bkg.getCoord()[2]);

			if (bkg.getCoord()[2] >= 0 ){} else {
			cout << "** There is no space in CCD for background selection for source: " << i << " in the list."<< endl;
			cout << "" << endl;
			}
		}

		// Insert Bkg RAWX,RAWY,CCDNR in Source List:
		// Add Column to existing FITS table
		ColumnData *bkg_rawx_ = src_tab->addColumn("RAWX", Dal::Real32)->data();
		real32 *bkg_rawx__ = bkg_rawx_->real32Data();
		ColumnData *bkg_rawy_ = src_tab->addColumn("RAWY", Dal::Real32)->data();
		real32 *bkg_rawy__ = bkg_rawy_->real32Data();
		ColumnData *bkg_ccd_ = src_tab->addColumn("CCDNR", Dal::Int8)->data();
		int8 *bkg_ccd__ = bkg_ccd_->int8Data();
		ColumnData *bkg_rad_ = src_tab->addColumn("BKG_RAD", Dal::Real32)->data();
		real32 *bkg_rad__ = bkg_rad_->real32Data();

			// Populate columns :
			for (unsigned long i = 0 ; i < src_N ; ++i){
				bkg_rawx__[i] = bkg_rawx[i];
				bkg_rawy__[i] = bkg_rawy[i];
				bkg_ccd__[i] = src_ccdnr[i]; // Bkg region in the same CCD as the source
				bkg_rad__[i] = bkg_rad[i];
			}

		dataSetServer->close(src_set);  // Close Table FITS file (clone) before 'edet2sky'

		//
		// Convert Detector to Sky Coordinates for Bkg
		//
		// Convert Bkg RAWX,RAWY,CCDNR -> Bkg Sky (RA,Dec and X,Y)
		// edet2sky(Bkg RAWX,RAWY,CCD) == Bkg RA, Dec and Bkg X,Y  --> 4 extra columns in input source list
		Info::Task* _edet2sky = task("edet2sky");
		ParameterInfo& edet2skypars = *_edet2sky->parameters();
		edet2skypars.revertToDefault();
		edet2skypars["calinfoset"]->setValue(_imagefile);
		edet2skypars["datastyle"]->setValue("set");
		edet2skypars["inputunit"]->setValue("raw"); // Optional. Default is 'raw'
		edet2skypars["intab"]->setValue(src_table_clone);

				  	try{
				   			task(edet2sky);
				  		}
				  		catch (SasException::Error e) {
				  			errstr << _edet2sky->name() << " terminated with the error " << e.code
				   			   << " that said: \n" << e.message
				   			   << Sas::warning("SubTaskError");
				   		}
				   		catch (SasException::Fatal e) {
				   			errstr << _edet2sky->name() << " terminated with the fatal error " << e.code
				   			   << " that said: \n" << e.message
				   			   << Sas::warning("SubTaskError");
				   		}

}

string ProcSrcList::getTempSrcList()
{
	return _res_tmpsrclist;
}

